import os
import re
import warnings
from typing import Dict, Optional, List, Tuple, Union

from bs4 import BeautifulSoup
import pandas as pd
import numpy as np
import us 

from genpeds.downloader import get_year_iter


VARIABLE_RENAME = {
    'characteristics' : {
        'unitid' : 'id', 'instnm' : 'name',
          'addr' : 'address', 'city' : 'city', 'stabbr' : 'state', 'zip' : 'zipcode', 
          'webaddr' : 'webaddress', 
          'longitud' : 'longitude', 'latitude' : 'latitude'
    },

    'admissions' : {
        'unitid' : 'id', 
        'applcnm' : 'men_applied', 'applcnw' : 'women_applied', 
        'admssnm' : 'men_admitted', 'admssnw' : 'women_admitted',
        'satpct' : 'share_submit_sat', 'actpct' : 'share_submit_act',                                      
        'satvr25' : 'sat_rw_25', 'satvr75' : 'sat_rw_75', 
        'satmt25' : 'sat_math_25', 'satmt75' : 'sat_math_75', 
        'actcm25' : 'act_comp_25', 'actcm75' : 'act_comp_75',
        'acten25' : 'act_eng_25', 'acten75' : 'act_eng_75', 
        'actmt25' : 'act_math_25', 'actmt75' : 'act_math_75',
        'enrlftm' : 'men_ft_enrolled', 'enrlftw' : 'women_ft_enrolled',
        'enrlptm' : 'men_pt_enrolled', 'enrlptw' : 'women_pt_enrolled',
        'enrlm' : 'men_enrolled', 'enrlw' : 'women_enrolled',
        'applcn' : 'tot_applied', 'admssn' : 'tot_admitted', 'enrlt' : 'tot_enrolled',
        'acten50' : 'act_eng_50', 'actmt50' : 'act_math_50', 'actcm50' : 'act_comp_50',
        'satvr50' : 'sat_rw_50', 'satmt50' : 'sat_math_50'
    },

    'enrollment' : {
        'unitid' : 'id', 'line' : 'line', 'efrace15' : 'totmen', 'efrace16' : 'totwomen',
        'eftotlm' : 'totmen', 'eftotlw' : 'totwomen', 'efwhitm' : 'wtmen', 'efwhitw' : 'wtwomen',
        'efbkaam' : 'bkmen', 'efbkaaw' : 'bkwomen', 'efhispm' : 'hspmen', 'efhispw' : 'hspwomen', 
        'efasiam' : 'asnmen', 'efasiaw' : 'asnwomen', 'eftotlw' : 'totwomen',
        'efrace11' : 'wtmen', 'efrace12' : 'wtwomen', 'efrace03' : 'bkmen', 'efrace04' : 'bkwomen', 
        'efrace09' : 'hspmen', 'efrace10' : 'hspwomen', 'efrace07' : 'asnmen', 'efrace08' : 'asnwomen'
    },

    'completion' : {
        'unitid' : 'id', 'cipcode' : 'cip', 'awlevel' : 'awlevel',
        'crace15' : 'totmen', 'crace16' : 'totwomen',
        'ctotalm' : 'totmen', 'ctotalw' : 'totwomen', 
        'cwhitm' : 'wtmen', 'cwhitw' : 'wtwomen', 'cbkaam' : 'bkmen', 'cbkaaw' : 'bkwomen', 
        'chispm' : 'hspmen', 'chispw' : 'hspwomen', 'casiam' : 'asnmen', 'casiaw' : 'asnwomen',
        'crace11' : 'wtmen', 'crace12' : 'wtwomen', 'crace03' : 'bkmen', 'crace04' : 'bkwomen', 
        'crace09' : 'hspmen', 'crace10' : 'hspwomen', 'crace07' : 'asnmen', 'crace08' : 'asnwomen'
    },

    'cip' : {
        # nothing needed for now.
    },

    'graduation' : {
        'grtotlm' : 'totmen', 'grtotlw' : 'totwomen', 
        'grwhitm' : 'wtmen', 'grwhitw' : 'wtwomen',
        'grbkaam' : 'bkmen', 'grbkaaw' : 'bkwomen', 
        'grhispm' : 'hspmen', 'grhispw' : 'hspwomen', 
        'grasiam' : 'asnmen', 'grasiaw' : 'asnwomen', 
        'grrace15' : 'totmen', 'grrace16' : 'totwomen',
        'grrace11' : 'wtmen', 'grrace12' : 'wtwomen', 
        'grrace03' : 'bkmen', 'grrace04' : 'bkwomen', 
        'grrace09' : 'hspmen', 'grrace10' : 'hspwomen', 
        'grrace07' : 'asnmen', 'grrace08' : 'asnwomen',
        'chrtstat' : 'chrtstat', 'section' : 'section', 
        'cohort' : 'cohort', 'unitid' : 'id', 'grtype' : 'grtype'
    }
}


def clean_characteristics(characteristics_dir: str = 'characteristicsdata',
                          year_range: Optional[Union[Tuple[int,int], List[int], int]] = None) -> pd.DataFrame:
    '''
    cleans institution characteristics data and returns complete characteristics data

    :param characteristics_dir: directory where raw enrollment data is located
    :year_range: range of years to clean data from. This is in case that you have more data than you want to actually clean and return
    '''
    warnings.filterwarnings('ignore', category=FutureWarning)
    sorted_files = sorted(os.listdir(characteristics_dir))
    rename_dict = VARIABLE_RENAME['characteristics']
    state_mappings = us.states.mapping('abbr', 'name') # get abbr -> names for states
    state_mappings['DC'] = 'District of Columbia' # add Washington D.C. to state mapping

    master_df = pd.DataFrame()

    dtypes = {
        'unitid' : str, 'instnm' : str, 'addr' : str, 'city' : str, 
        'stabbr' : str, 'zip' : str, 'webaddr' : str, 'longitud' : str, 'latitude' : str
    }
    for file in sorted_files:
        file_path = os.path.join(characteristics_dir, file)
        year_num = re.split(r'_|\.', f'{file}')[1]
        
        if year_range:
            year_iter = get_year_iter(subject='characteristics',
                                      year_range=year_range)
            if int(year_num) not in year_iter:
                continue

        df = pd.read_csv(file_path, dtype=dtypes, encoding_errors='replace', low_memory=False)
        df = df.rename(str.lower, axis='columns')
        
        if int(year_num) > 1998:
            if int(year_num) > 2008:
                filt_col = ['unitid', 'instnm', 'addr', 'city', 'stabbr', 'zip', 'webaddr', 'longitud', 'latitude']
            else:
                filt_col = ['unitid', 'instnm', 'addr', 'city', 'stabbr', 'zip', 'webaddr'] # long/lat NA for before 2008
        else:
            filt_col = ['unitid', 'instnm', 'addr', 'city', 'stabbr', 'zip'] # for years < 1999
        df_filtered = df.loc[:, filt_col]
        
        for col in ['instnm', 'addr', 'city']:
            df_filtered[col] = df_filtered[col].str.title() # TitleCase
        df_filtered['year'] = int(year_num) # year identifier
        df_filtered['unitid'] = df_filtered['unitid'].astype(str).str.strip() # make id into string
        df_filtered['stabbr'] = df_filtered['stabbr'].map(state_mappings) # state abbreviation to state name
        master_df = pd.concat([master_df, df_filtered], ignore_index=True)
    
    master_df = master_df.rename(columns=rename_dict) # rename vars
    return master_df


def clean_admissions(admissions_dir: str = 'admissionsdata',
                     year_range: Optional[Union[Tuple[int,int], List[int], int]] = None) -> pd.DataFrame:
    '''
    cleans yearly admissions data and returns complete admissions data
    
    :param admissions_dir: directory where raw admissions data is located
    :year_range: range of years to clean data from. This is in case that you have more data than you want to actually clean and return
    '''
    warnings.filterwarnings('ignore', category=FutureWarning)
    sorted_files = sorted(os.listdir(admissions_dir)) # unnecessary, but helps with error checking
    rename_dict = VARIABLE_RENAME['admissions']

    master_df = pd.DataFrame()
    
    for file in sorted_files:
        file_path = os.path.join(admissions_dir, file)
        year_num = re.split(r'_|\.', f'{file}')[1]
        
        if year_range:
            year_iter = get_year_iter(subject='admissions',
                                      year_range=year_range)
            if int(year_num) not in year_iter:
                continue

        df = pd.read_csv(file_path, dtype=str) # read in df
        df = df.rename(str.lower, axis='columns') # some df's have all uppercase, some have all lowercase
        df.columns = df.columns.str.strip() # some column names have right spaces
        
        cols_to_filter = [col for col in rename_dict.keys() if col in df.columns] # cols to filter per year
        df_filtered = df.reindex(columns=cols_to_filter)
        df_filtered = df_filtered.rename(columns=rename_dict) # rename cols

        for col in df_filtered.columns:
            if col == 'id':
                df_filtered[col] = df_filtered[col].astype(str).str.strip() # id str
            else:
                df_filtered[col] = pd.to_numeric(df_filtered[col], errors='coerce')

        if int(year_num) == 2001:
            df_filtered['men_enrolled'] = df_filtered['men_ft_enrolled'] + df_filtered['men_pt_enrolled']
            df_filtered['women_enrolled'] = df_filtered['women_ft_enrolled'] + df_filtered['women_pt_enrolled']
        if 'tot_applied' not in df_filtered.columns:
            df_filtered['tot_applied'] = df_filtered['men_applied'] + df_filtered['women_applied']
            df_filtered['tot_admitted'] = df_filtered['men_admitted'] + df_filtered['women_admitted']
            df_filtered['tot_enrolled'] = df_filtered['men_enrolled'] + df_filtered['women_enrolled']

        for i in ['men', 'women']:
            df_filtered[f'accept_rate_{i}'] = np.where(
            df_filtered[f'{i}_applied'] == 0,
            np.nan,
            (df_filtered[f'{i}_admitted'] / df_filtered[f'{i}_applied'] * 100)
            )
    
            df_filtered[f'yield_rate_{i}'] = np.where(
            df_filtered[f'{i}_admitted'] == 0,
            np.nan,
            (df_filtered[f'{i}_enrolled'] / df_filtered[f'{i}_admitted'] * 100)
            )
        
        df_filtered['year'] = int(year_num) # year identifier
        df_filtered['men_applied_share'] = df_filtered['men_applied'] / df_filtered['tot_applied'] * 100
        df_filtered['men_admitted_share'] = df_filtered['men_admitted'] / df_filtered['tot_admitted'] * 100

        master_df = pd.concat([master_df, df_filtered], ignore_index=True)
    
    # unneeded columns
    admissions_df = master_df.drop(columns=['women_applied', 'women_admitted', 'women_enrolled',
                                            'men_ft_enrolled', 'men_pt_enrolled', 'women_ft_enrolled', 'women_pt_enrolled'],
                                   errors='ignore')

    return admissions_df


def clean_enrollment(enrollment_dir: str = 'enrollmentdata', 
                     student_level: str = 'undergrad',
                     year_range: Optional[Union[Tuple[int,int], List[int], int]] = None) -> pd.DataFrame:
    '''
    cleans yearly enrollment data and returns complete student enrollment data

    :enrollment_dir: directory where raw enrollment data is located
    :student_level: level of enrollment; options include ['undergrad', 'grad']
    :year_range: range of years to clean data from. This is in case that you have more data than you want to actually clean and return
    '''
    warnings.filterwarnings('ignore', category=FutureWarning)
    sorted_files = sorted(os.listdir(enrollment_dir)) # unnecessary, but helps with error checking
    rename_dict = VARIABLE_RENAME['enrollment']

    grad_rules = [
        (lambda y: y in [1984,1985], 'line in [11,25,10,24]'),
        (lambda y: (y == 1986) or (y in range(1990,1999)), 'line in [14,28,9,10,23,24]'),
        (lambda y: y in [1987,1988,1989], 'line in [14,28]'),
        (lambda y: y == 1999, 'line in [32,52,16]'),
        (lambda y:  y in range(2000,2009), 'line in [11,25,9,23]'),
        (lambda y: y in range(2009,2025), 'line in [11,25]')
    ]
    
    df_list = []  

    for file in sorted_files:
        file_path = os.path.join(enrollment_dir, file)
        year_num = re.split(r'_|\.', f'{file}')[1]
        
        if year_range:
            year_iter = get_year_iter(subject='enrollment',
                                      year_range=year_range)
            if int(year_num) not in year_iter:
                continue

        df = pd.read_csv(file_path, dtype=str) # read in df
        df = df.rename(str.lower, axis='columns') # some df's have all uppercase, some have all lowercase

        if all(col in df.columns for col in ['efrace10', 'eftotlm']):
            cols_to_filter = [col for col in rename_dict.keys() if 
                              (col in df.columns) and ('efrace' not in col)] # annoying thing with duplicate cols in 
                                                                                # some years
        else:
            cols_to_filter = [col for col in rename_dict.keys() if col in df.columns]
            

        df_filtered = df.reindex(columns=cols_to_filter)
        df_filtered = df_filtered.rename(columns=rename_dict) # rename cols

        for col in df_filtered.columns:
            if col == 'id':
                df_filtered[col] = df_filtered[col].astype(str).str.strip() # id identifier
            else:
                df_filtered[col] = pd.to_numeric(df_filtered[col], errors='coerce')

        if student_level == 'undergrad':
            if int(year_num) < 1986:
                student_query = 'line == 1 or line == 15' # captures total full-time and total part-time undergrads, respectively
            else:
                student_query = 'line == 8 or line == 22' 
        elif student_level == 'grad':
            for cond,frmt in grad_rules:
                if cond(int(year_num)):
                    student_query = frmt # captures full-time and part-time graduate and first-professional students
                    break
            else:
                raise ValueError(f'No formatted rule for year {year_num}')
        else:
            raise ValueError("student_level must be 'undergrad' or 'grad' ")

        students = df_filtered.query(student_query) # filter data to total students
        
        if 'wtmen' not in students.columns:
            cols_to_sum = ['totmen', 'totwomen']
        else:
            cols_to_sum = ['totmen', 'totwomen', 'wtmen', 'wtwomen','bkmen', 'bkwomen','hspmen', 'hspwomen','asnmen', 'asnwomen']
        
        students_by_inst = students.groupby('id')[cols_to_sum].sum() # sum full-time and part-time students by school

        # Calculate male student share without using eval
        students_by_inst['totmen_share'] = students_by_inst['totmen'] / (students_by_inst['totmen'] + students_by_inst['totwomen']) * 100 # male student share

        # Resetting index here to be compatible with the previous version,
        # but I'm not sure why it's necessary.
        students_by_inst = students_by_inst.reset_index()

        students_by_inst['year'] = int(year_num) # get year marker for each set
        students_by_inst['studentlevel'] = student_level # get student level identifier

        if 'wtmen' in students_by_inst.columns:
            for attr in ['wt', 'bk', 'hsp', 'asn']:
                # Calculate race share breakdowns without using eval
                students_by_inst[f'tot{attr}_share'] = (students_by_inst[f'{attr}men'] + students_by_inst[f'{attr}women']) / (students_by_inst['totmen'] + students_by_inst['totwomen']) * 100

        df_list.append(students_by_inst)

    
    master_df = pd.concat(df_list, ignore_index=True)

    return master_df


def clean_completion(completion_dir: str = 'completiondata', 
                     level: str = 'bach',
                     year_range: Optional[Union[Tuple[int,int], List[int], int]] = None) -> pd.DataFrame:
    '''
    cleans yearly completion data and returns complete completions data

    :param completion_dir: directory where raw completion data is located
    :param level: level of degree, options include ['assc', 'bach', 'mast', 'doct']
    :year_range: range of years to clean data from. This is in case that you have more data than you want to actually clean and return
    '''
    warnings.filterwarnings('ignore', category=FutureWarning)
    sorted_files = sorted(os.listdir(completion_dir)) # unnecessary, but helps with error checking
    rename_dict = VARIABLE_RENAME['completion']

    deglevel_rules = [
        (lambda l,y: l == 'assc', 'awlevel == 3'),
        (lambda l,y: l == 'bach', 'awlevel == 5'),
        (lambda l,y: l == 'mast', 'awlevel == 7'),
        (lambda l,y: (l == 'doct') and (y < 2010), 'awlevel == 9'),
        (lambda l,y: (l == 'doct') and (y >= 2010), 'awlevel >= 17 and awlevel <= 19')
    ]
    
    master_df = pd.DataFrame()

    for file in sorted_files:
        file_path = os.path.join(completion_dir, file)
        year_num = re.split(r'_|\.', f'{file}')[1]
        
        if year_range:
            year_iter = get_year_iter(subject='completion',
                                      year_range=year_range)
            if int(year_num) not in year_iter:
                continue
        
        df = pd.read_csv(file_path, dtype=str) # read in df
        df = df.rename(str.lower, axis='columns') # some df's have all uppercase, some have all lowercase
        
        if all(col in df.columns for col in ['crace10', 'ctotalm']):
            cols_to_filter = [col for col in rename_dict.keys() if 
                              (col in df.columns) and ('crace' not in col)] # annoying thing with duplicate cols in 
                                                                                # some years
        else:
            cols_to_filter = [col for col in rename_dict.keys() if col in df.columns]

        df_filtered = df.reindex(columns=cols_to_filter)
        df_filtered = df_filtered.rename(columns=rename_dict)
        for col in df_filtered:
            if col not in ['id', 'cip']:
                df_filtered[col] = pd.to_numeric(df_filtered[col], errors='coerce')
        
        for cond,frmt in deglevel_rules:
            if cond(level, int(year_num)):
                level_query = frmt
                break
        else:
            raise ValueError("level must be 'assc', 'bach', 'mast' or 'doct'") 
        
        completions = df_filtered.query(level_query)
        race_cols = [col for col in completions.columns if 'men' in col] # race columns to group
        completions = completions.groupby(['id', 'cip'])[race_cols].sum().reset_index()

        completions = completions.eval('totmen_share = totmen / (totmen + totwomen) * 100') # maleshare within each major
        if 'wtmen' in completions.columns:
            for attr in ['wt', 'bk', 'hsp', 'asn']:
                eval_str = f'tot{attr}_share = ({attr}men + {attr}women) / (totmen + totwomen) * 100' # race share breakdowns
                completions = completions.eval(eval_str)
        
        completions['deglevel'] = level # adds level identifier
        completions['year'] = int(year_num) # adds year identifier
        completions['cip'] = completions['cip'].astype(str).str.strip()
        
        master_df = pd.concat([master_df, completions], ignore_index=True)

    return master_df


def clean_cip_html(file_path: str) -> Dict[str,str]:
    '''
    returns dict of CIP subject code:label pairs for a given year's CIP dictionary html.
    
    :param file_path: string path to CIP data dictionary html
    :year_range: range of years to clean data from. This is in case that you have more data than you want to actually clean and return
    '''
    with open(file_path) as filehandle:
        soup = BeautifulSoup(filehandle, 'html.parser')
        rows = soup.find_all('tr', attrs={'bgcolor': ['White', 'Silver']})
        dat_rows = rows[1:]
        label_dict = {}
        for row in dat_rows:
            cells = row.find_all('td')
            if len(cells) > 1:
                label = cells[0].text.strip()
                val = cells[1].text.strip()
                if label == 'Totals':
                    break # end of relevant table
                else:
                    label_dict[val] = label
        return label_dict


def clean_cip(cip_codes_dir: str = 'cipdata',
              year_range: Optional[Union[Tuple[int,int], List[int], int]] = None) -> pd.DataFrame:
    '''
    cleans yearly CIP data and returns full dataframe

    :param cip_codes_dir: directory where raw CIP data is located
    :year_range: range of years to clean data from. This is in case that you have more data than you want to actually clean and return
    '''
    warnings.filterwarnings('ignore', category=FutureWarning)
    warnings.filterwarnings('ignore', category=UserWarning)
    sorted_files = sorted(os.listdir(cip_codes_dir))
    master_df = pd.DataFrame()

    for file in sorted_files:
        file_path = os.path.join(cip_codes_dir, file)
        year_num = re.split(r'_|\.', f'{file}')[1]
        
        if year_range:
            year_iter = get_year_iter(subject='cip',
                                      year_range=year_range)
            if int(year_num) not in year_iter:
                continue
        
        ext = re.split(r'_|\.', f'{file}')[2]
        if ext == 'html':
            html_dict = clean_cip_html(file_path=file_path)
            df = pd.DataFrame({'cip_description' : html_dict.values(),
                               'cip' : html_dict.keys()}, dtype=str)
        else:
            df = pd.read_excel(file_path, sheet_name='Frequencies', dtype=str)
            df = df.rename(columns=str.lower)
            df = df.query('varname == "CIPCODE" or varname == "Cipcode"').loc[:, ['codevalue', 'valuelabel']]
            df = df.rename(columns={'codevalue' : 'cip', 'valuelabel' : 'cip_description'})
        df['year'] = int(year_num) # year identifier
        
        master_df = pd.concat([master_df, df], ignore_index=True)
        master_df['cip'] = master_df['cip'].astype(str).str.strip() # strip spaces
        master_df['cip_description'] = master_df['cip_description'].str.title().replace(r'^(\d+)\s-\s', '', regex=True)

    return master_df


def clean_graduation(graduation_dir: str = 'graduationdata', 
                     deg_level: str = 'bach',
                     year_range: Optional[Union[Tuple[int,int], List[int], int]] = None) -> pd.DataFrame:
    '''
    cleans yearly graduation data and returns complete graduation data

    :graduation_dir: directory where raw completion data is located
    :deg_level: degree level; options include ['assc', 'bach']
    :year_range: range of years to clean data from. This is in case that you have more data than you want to actually clean and return
    '''
    warnings.filterwarnings('ignore', category=FutureWarning)
    sorted_files = sorted(os.listdir(graduation_dir)) # unnecessary, but helps with error checking
    rename_dict = VARIABLE_RENAME['graduation']
    master_df = pd.DataFrame()

    for file in sorted_files:
        file_path = os.path.join(graduation_dir, file)
        year_num = re.split(r'_|\.', f'{file}')[1]
        
        if year_range:
            year_iter = get_year_iter(subject='graduation',
                                      year_range=year_range)
            if int(year_num) not in year_iter:
                continue
        
        df = pd.read_csv(file_path, dtype=str) # read in df
        df = df.rename(str.lower, axis='columns') # some df's have all uppercase, some have all lowercase

        if all(col in df.columns for col in ['grrace10', 'grtotlm']):
            cols_to_filter = [col for col in rename_dict.keys() if 
                              (col in df.columns) and ('grrace' not in col)] # annoying thing with duplicate cols in 
                                                                                # some years
        else:
            cols_to_filter = [col for col in rename_dict.keys() if col in df.columns]
        df_filtered = df.reindex(columns=cols_to_filter)
        df_filtered = df_filtered.rename(columns=rename_dict)
        
        for col in df_filtered.columns:
            if col != 'id':
                df_filtered[col] = pd.to_numeric(df_filtered[col], errors='coerce') # convert cols to float
        
        if deg_level == 'bach':
            deg_query = '(8 <= grtype <= 9) and (12 <= chrtstat <= 13) and section == 2'
            denom = 8
            num = 9
        elif deg_level == 'assc':
            deg_query = '(29 <= grtype <= 30) and (12 <= chrtstat <= 13) and section == 4'
            denom = 29
            num = 30
        else:
            raise ValueError("deg_level must be 'assc', 'bach'")

        grads = df_filtered.query(deg_query)
        pivoted_grads = grads.pivot(index='id', columns='grtype', values=['totmen', 'totwomen', # get cohort and grads
                                                                                        'wtmen', 'wtwomen',
                                                                                        'bkmen', 'bkwomen',
                                                                                        'hspmen', 'hspwomen',
                                                                                        'asnmen', 'asnwomen'])
        for i in ['tot', 'wt', 'bk', 'hsp', 'asn']:
            pivoted_grads[f'gradrate_{i}men'] = (pivoted_grads[f'{i}men'][num] /    
                                                     pivoted_grads[f'{i}men'][denom] * 100)
            
            pivoted_grads[f'gradrate_{i}women'] = (pivoted_grads[f'{i}women'][num] / 
                                                     pivoted_grads[f'{i}women'][denom] * 100) # grad rates
        pivoted_grads = pivoted_grads.reset_index()

        rnm_columns = pivoted_grads.columns.droplevel(1).tolist() # renaming columns
        for i in range(len(rnm_columns)):
            if rnm_columns[i] == rnm_columns[i - 1]:
                rnm_columns[i] = f'{rnm_columns[i]}_graduated'
        pivoted_grads.columns = rnm_columns
        
        pivoted_grads['year'] = int(year_num) # get year identifiers
        pivoted_grads['deglevel'] = deg_level

        master_df = pd.concat([master_df, pivoted_grads], ignore_index=True)
    
    return master_df
        

CLEANERS = {
    'characteristics' : clean_characteristics,
    'admissions' : clean_admissions,
    'enrollment' : clean_enrollment,
    'completion' : clean_completion,
    'cip' : clean_cip,
    'graduation' : clean_graduation
}



import os

import pytest
import pandas as pd


from genpeds import Characteristics, Admissions, Enrollment, Completion, Graduation

# classes to test
# - Characteristics
# - Admissions
# - Enrollment
# - Completion (CIP automatically gets tested here, depending on args)
# - Graduation

# non-unique methods to test
# - .get_description()
# - .get_available_years()
# - .get_available_vars()
# - .lookup_var()
# - .scrape()

# unique methods to test
# .clean()
# .run()


def get_iter_range(years):
    '''returns range of years depending on object type.'''
    if isinstance(years, tuple):
        start,end = years
        iter_range = range(start, end + 1)
    elif isinstance(years, list):
        iter_range = years
    elif isinstance(years, int):
        iter_range = [years]
    else:
        raise ValueError('years must be tuple, list or int.')
    return iter_range

@pytest.mark.parametrize('subject_name, subject, year_range', [
    ('characteristics', Characteristics, (2000,2005)), 
    ('admissions', Admissions, [2005,2009,2012,2022]),
    ('enrollment', Enrollment, [1984,1989,1990,2003,2024]),
    ('completion', Completion, (2002,2012)),
    ('graduation', Graduation, (2015,2019))
])
def test_common_methods(subject_name, subject, year_range):
    '''test non-unique methods for genpeds classes'''
    genpeds_class = subject(year_range=year_range)
    # test if .year_range() returns correct ranges
    if isinstance(genpeds_class, (Characteristics, Enrollment, Completion)):
        assert genpeds_class.get_available_years() == (1984,2024) # check if returns appropriate range
    elif isinstance(genpeds_class, Admissions):
        assert genpeds_class.get_available_years() == (2001,2024)
    else:
        assert genpeds_class.get_available_years() == (2000,2024)
    # test if correct description is returned, lazy test but it's fine for now
    assert subject_name in genpeds_class.get_description().lower()
    # test .available_vars()
    assert isinstance(genpeds_class.get_available_vars(), dict)
    assert 'year' in genpeds_class.get_available_vars().keys()
    # test .lookup_var()
    assert 'Institutional unique identifier ID' in  genpeds_class.lookup_var('id') 
    # test .scrape() works, and that all files are downloaded in relevant dirs
    genpeds_class.scrape()
    dir_name = f'{subject_name}data' 
    assert os.path.exists(dir_name) # was subject directory created?
    for yr in get_iter_range(years=year_range):
        f_name = f'{subject_name}_{yr}.csv'
        assert f_name in os.listdir(dir_name) # was each year's file downloaded?

# test .clean() and .run() for subjects that don't take additional args
# e.g., Characteristics and Admissions
@pytest.mark.parametrize('subject, year_range', [
    (Characteristics, [2000,2005, 2010]),
    (Admissions, [2010, 2013, 2017])
])
def test_cleanrun_basic(subject, year_range):
    '''test .clean() and .run() for Characteristics and Admissions.'''
    # test clean first, assuming we still have data in our directory from before
    df1 = subject().clean()

    assert isinstance(df1, pd.DataFrame)
    assert 'id' in df1.columns
    if subject == Characteristics:
        assert 'name' in df1.columns
        assert len(df1.loc[df1['name'] == 'Harvard University']) > 0
        assert df1.loc[df1['name'] == 'Harvard University', 'state'].mode()[0] == 'Massachusetts'
    if subject == Admissions:
        assert 16 < df1['act_eng_25'].mean() < 36
        assert 16 < df1['act_math_25'].mean() < 36

    # test run now
    df2 = subject(year_range=year_range).run()

    assert isinstance(df2, pd.DataFrame)
    if subject == Characteristics:
        assert len(df1.loc[df1['name'] == 'Vanderbilt University']) > 0
        assert df1.loc[df1['name'] == 'Vanderbilt University', 'state'].mode()[0] == 'Tennessee'
    if subject == Admissions:
        assert 16 < df1['act_eng_75'].mean() < 36
        assert 16 < df1['act_math_75'].mean() < 36

# test Enrollment().clean() and Enrollment().run() now
# .clean() and .run() take in a 'student_level' param
# with 'undergrad' and 'grad' as the options
# here, we'll also see if the merge_with_char param works
@pytest.mark.parametrize('lev, year_range', [
    ('undergrad', (1985,1990)),
    ('grad', [1987, 2004, 2012, 2022])
])
def test_enrollment_extended(lev, year_range):
    '''tests Enrollment().clean() and Enrollment().run()'''
    # test Enrollment.clean()
    df1 = Enrollment().clean(student_level=lev) # try both 'undergrad' and 'grad'

    assert isinstance(df1, pd.DataFrame)
    for col in ['totmen_share', 'totwt_share', 
                'totbk_share', 'tothsp_share', 'totasn_share']:
        assert 0 <= df1.loc[df1['id'] == '190150', col].mean() <= 100 
                                           # some may be out of range

    # test Enrollment.run(), with merge_with_char
    df2 = Enrollment(year_range=year_range).run(merge_with_char=True)

    assert 'name' in df2.columns
    assert len(df2.loc[df2['name'] == 'Stanford University']) > 0

# Completion().clean() and Completion().run()
# these methods take the degree_level param 
# options include ['assc', 'bach', 'mast', 'doct']
# run also takes get_cip_codes bool param
@pytest.mark.parametrize('lev, year_range', [
    ('assc', [2012,2022]),
    ('bach', [1990,1994,2000]),
    ('mast', [2015,2016,2019]),
    ('doct', [1999, 2020,2024])
])
def test_completion_extended(lev, year_range):
    '''tests Completion().clean() and Completion().run()'''
    # test Completion.clean()
    df1 = Completion().clean(degree_level=lev) # try all degree levels

    assert isinstance(df1, pd.DataFrame)
    assert 0 <= df1['totmen_share'].mean() <= 100

    # test Completion.run(), with merge_with_char and get_cip_codes
    df2 = Completion(year_range=year_range).run(merge_with_char=True, get_cip_codes=True)

    assert 'name' in df2.columns
    assert len(df2.loc[df2['name'] == 'Stanford University']) > 0
    assert 'cip_description' in df2.columns
    assert len(df2.loc[df2['cip_description'] == 'Physics, General']) > 0
    
# test Graduation.clean() and Graduation.run()
# Graduation takes degree_level param
# options include 'assc', 'bach'
@pytest.mark.parametrize('lev, year_range', [
    ('assc', [2004,2005,2018,2020]),
    ('bach', [2017,2019,2022,2024])
])
def test_graduation_extended(lev, year_range):
    '''tests Graduation.clean() and Graduation.run()'''
    # Graduation.clean()
    df1 = Graduation().clean(degree_level=lev)
    
    assert isinstance(df1, pd.DataFrame)
    assert 0 <= df1['gradrate_totmen'].mean() <= 100
    assert 0 <= df1['gradrate_totwomen'].mean() <= 100

    # Graduation.run()
    df2 = Graduation(year_range=year_range).run(degree_level=lev, merge_with_char=True)

    assert 'name' in df2.columns # check if Characteristics merge works
    for col in ['gradrate_totmen', 'gradrate_totwomen']:
        for school in ['Carnegie Mellon University', 'Cornell University', 
                                'Princeton University', 'Brown University']:
            if len(df2.loc[df2['name'] == school]) > 0: # these schools don't have assc programs
                assert df2.loc[df2['name'] == school, 
                                             col].mean() > 80 # just a hunch




import os
import shutil

import pytest

from genpeds import scrape_ipeds_data

@pytest.mark.parametrize('subject, year_range_char', [
    # characteristics
    ('characteristics', (1990,2003)),
    ('characteristics', [2002,2004,2006,2008]),
    ('characteristics', 2024),
    # admissions
    ('admissions', (2008,2012)),
    ('admissions', [2001,2014,2018]),
    ('admissions', 2022),
    # enrollment
    ('enrollment', [2018,2020,2024]),
    ('enrollment', (1985,1990)),
    ('enrollment', 2002),
    # completion
    ('completion', (1984,1989)),
    ('completion', [2003,2015,2024]),
    ('completion', 2010),
    # graduation
    ('graduation', (2000,2005)),
    ('graduation', [2008,2013,2017,2019]),
    ('graduation', 2022)

])

def test_subject_downloads(subject, year_range_char):
    '''test IPEDS variable-subject data scraping'''
    download_dir = f'{subject}data'
    if os.path.exists(download_dir):
        shutil.rmtree(download_dir)

    try:
        scrape_ipeds_data(subject=subject, year_range=year_range_char)
        assert os.path.exists(download_dir), f'{subject.title()} directory not found.'
        # tuple range
        if isinstance(year_range_char, tuple):
            start,end = year_range_char
            for yr in range(start, end + 1):
                f_name = f'{subject}_{yr}.csv'
                assert f_name in os.listdir(download_dir), f'File {f_name} not found.'
        # list range
        elif isinstance(year_range_char, list):
            for yr in year_range_char:
                f_name = f'{subject}_{yr}.csv'
                assert f_name in os.listdir(download_dir), f'File {f_name} not found.'
        # single year
        else:
            f_name = f'{subject}_{year_range_char}.csv'
            assert f_name in os.listdir(download_dir), f'File {f_name} not found.'
    finally:
        shutil.rmtree(download_dir)

# you're reading this? 
# well, you should be reading 'The Master and Margarita' by Mikhail Bulgakov instead.
# you won't regret it friend :)


import concurrent.futures
import time
import random
import zipfile
import os
from pathlib import Path
import json
import warnings
import re
from typing import (Optional, 
                    Union, 
                    List, 
                    Tuple,
                    Dict)

import requests


def get_year_iter(subject: str,
                  year_range: Optional[Union[Tuple[int,int], List[int], int]] = None) -> List[int]:
    '''
    get year iterable based on subject and year input

    :param subject: string identifying which subject data to download. The subjects available are:
     ['characteristics', 'admissions', 'enrollment', 'completion', 'cip', 'graduation']
    
    :param year_range: tuple of year integers (indicates a range), iterable of year integers (indicates group of individual years), or single year to pull data from. Data for 'characteristics', 'enrollment' and 'completion' are available for years 1984-2024, while 'graduation' is available for years 2000-2024. Defaults to all available years for a subject.
    '''
    subject = subject.lower()

    if not year_range:
        if subject == 'graduation':
            start, end = 2000, 2024
            iter_range = list(range(start, end + 1))  # default for graduation data
        elif subject == 'admissions':
            start, end = 2001, 2024
            iter_range = list(range(start, end + 1))  # default for admissions data
        else:
            start, end = 1984, 2024
            iter_range = list(range(start, end + 1))  # default for all other data
    else:
        if isinstance(year_range, tuple):
            start, end = year_range
            iter_range = list(range(start, end + 1))  # tuple follows regular range
        elif isinstance(year_range, list):
            iter_range = year_range  # list remains list
        elif isinstance(year_range, int):
            start = year_range
            iter_range = [start]  # integer becomes one-element list
        else:
            raise TypeError('Please enter a tuple range, list of integers, or a single integer')
    
    return iter_range


def get_file_endpoint(subject: str, 
                      year: int,
                      cfg: Dict[str, str]) -> Optional[str]:
    '''
    returns endpoint for a given subject in a given year.
    
    :param year: year for file; available years vary by subject.
    :param subject: subject.
    :param cfg: dict with subject-year endpoints
    '''
    try:
        endpoint = cfg[subject]["endpoints"][str(year)]
    except KeyError:
        warnings.warn(f'No endpoint for year {year}')
        return None
    return endpoint


def download_a_file(subject: str, 
                    year: int,
                    cfg: Dict[str,str]) -> Optional[str]:
    '''
    downloads an IPEDS subject-year data file.

    :param year: year for file; available years vary by subject.
    :param subject: subject.
    :param cfg: dict with subject-year endpoints
    '''
    dir = f'{subject}data'  # directory subject name
    prefix = subject    # file subject prefix

    if subject != 'cip':
        url_template = 'https://nces.ed.gov/ipeds/datacenter/data/{}.zip'
    else:
        url_template = 'https://nces.ed.gov/ipeds/datacenter/data/{}_Dict.zip'
    
    endpoint = get_file_endpoint(subject, year, cfg) 
    endpoint_url = url_template.format(endpoint)

    # for some reason, some 2024 files needs different url
    if year == 2024 and subject in ['admissions', 'enrollment', 'graduation']:
        endpoint_url = f'https://nces.ed.gov/ipeds/data-generator?year={year}&tableName={endpoint}'

    try:
        r = requests.get(endpoint_url)  
    except requests.HTTPError as er:
        return f"Year {year}: Error - {str(er)}"
        
    if '404 - File or directory not found' in r.text:
        return f"Year {year}: 404 - File not found"
    
    zipped_file = os.path.join(dir, f'{prefix}_{year}.zip')
    
    with open(zipped_file, 'wb') as f:
        f.write(r.content)

    with zipfile.ZipFile(zipped_file, 'r') as zfile:
        # there will at least be one file in each zip folder
        # to not have to deal with varying extensions, case, etc.
        # just pull the last file (in lex. order)
        # this way, we can also get the revised files instead of OG ones (they have '^.*_rv' in them)
        file_to_extract = sorted(zfile.namelist())[-1]
        zfile.extract(file_to_extract, dir)

        old = os.path.join(dir, file_to_extract)
        new = os.path.join(dir, re.sub(r'^.*\.', f'{prefix}_{year}.', old))

        os.rename(old, new)
        os.remove(zipped_file)
    
    return f'IPEDS {subject.title()} ({year}) successfully downloaded and extracted'


def scrape_ipeds_data(subject: str = 'characteristics', 
                      year_range: Optional[Union[Tuple[int,int], List[int], int]] = None, 
                      see_progress: bool = True) -> None:
    '''
    downloads NCES IPEDS data on specified years for a defined subject.
    
    :param subject: string identifying which subject data to download. The subjects available are:
     ['characteristics', 'admissions', 'enrollment', 'completion', 'cip', 'graduation']
    
    :param year_range: tuple of year integers (indicates a range), iterable of year integers (indicates group of individual years), or single year to pull data from. Data for 'characteristics', 'enrollment' and 'completion' are available for years 1984-2024, while 'graduation' is available for years 2000-2024. Defaults to all available years for a subject.

    :param see_progress: boolean that, when true, prints completion statement for extraction of each year. If false, no messages printed.
    
    ## available data

    - :characteristics: institutional characteristics, like a school's name, address. Certain variables, like a school's longitude and latitude are only available in later years. Available for years 1984-2024.

    - :admissions: Admissions data, like number of applications and acceptances by gender. Available for years 2001-2024.

    - :enrollment: fall enrollment by gender and institutional level (e.g., 4-year undergraduate program), with most years including enrollment by race and gender. Available for years 1984-2024.

    - :completion: completion of degrees by gender, level of degree and subject field (e.g., Bachelor's in Economics), with most years including completion by race and gender. Available for years 1984-2024.

    - :cip: CIP, or Classification of Instructional Programs, are key-value pairs for subject study fields. CIP's vary by year, and are relevant to identify subject field in completion data. Available for years 1984-2024.

    - :graduation: number of cohorts and graduates by gender, institutional level and graduation measure (e.g., students earning a bachelor's degree within 6 years of entering). Available for years 2000-2024.
    '''
    dir = f'{subject}data'
    prefix = subject
    # Determine the years to download
    iter_range = get_year_iter(subject=subject,
                               year_range=year_range)
    
    # check if data already downloaded, if so, drop from download list
    if os.path.isdir(dir): 
        iter_range2 = []
        stripped_list = [
            re.sub(r'\.(csv|html|xlsx|xls)','',ff) 
            for ff 
            in sorted(os.listdir(dir))
        ]
        for yr in iter_range:
            if f'{prefix}_{yr}' not in stripped_list:
                iter_range2.append(yr)
        iter_range = iter_range2
    else:
        os.makedirs(dir, exist_ok=True) # create subject dir

    # open endpoint cfg
    pkg_dir = Path(__file__).parent
    endpoint_path = pkg_dir / 'cfg.json'
    with open(endpoint_path, 'r') as cfgjf:
        cfg = json.load(cfgjf)

    # multithread to speed up the process
    with concurrent.futures.ThreadPoolExecutor(max_workers=7) as exec:
        future_to_year = {exec.submit(download_a_file, subject, year, cfg): year for year in iter_range}
        for future in concurrent.futures.as_completed(future_to_year):
            yr = future_to_year[future]
            try:
                result = future.result()
                if see_progress:
                    print(result) # if you want to see the progress
                time.sleep(random.uniform(0.05, 0.2)) # you're welcome NCES :)
            except Exception as exc:
                print(f"Year {yr} generated an exception: {exc}")
